@@ -1,5 +1,6 @@
+# MIT License
 #
-# Copyright (c) 2001 - 2019 The SCons Foundation
+# Copyright The SCons Foundation
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -19,33 +20,28 @@
 # LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-#
-
-__revision__ = "/src/script/src/engine/SCons/Script/SConsOptions.py none 2019-12-16 19:06:27 none"
 
+import gettext
 import optparse
 import re
 import sys
 import textwrap
 
-no_hyphen_re = re.compile(r'(\s+|(?<=[\w\!\"\'\&\.\,\?])-{2,}(?=\w))')
-
-try:
-    from gettext import gettext
-except ImportError:
-    def gettext(message):
-        return message
-_ = gettext
-
 import SCons.Node.FS
 import SCons.Platform.virtualenv
 import SCons.Warnings
+from . import Main
 
-OptionValueError        = optparse.OptionValueError
-SUPPRESS_HELP           = optparse.SUPPRESS_HELP
+no_hyphen_re = re.compile(r'(\s+|(?<=[\w!\"\'&.,?])-{2,}(?=\w))')
+_ = gettext.gettext
+OptionValueError = optparse.OptionValueError
+SUPPRESS_HELP = optparse.SUPPRESS_HELP
 
 diskcheck_all = SCons.Node.FS.diskcheck_types()
 
+experimental_features = {'warp_speed', 'transporter', 'ninja'}
+
+
 def diskcheck_convert(value):
     if value is None:
         return []
@@ -99,8 +95,8 @@ class SConsValues(optparse.Values):
     """
 
     def __init__(self, defaults):
-        self.__dict__['__defaults__'] = defaults
-        self.__dict__['__SConscript_settings__'] = {}
+        self.__defaults__ = defaults
+        self.__SConscript_settings__ = {}
 
     def __getattr__(self, attr):
         """
@@ -126,49 +122,73 @@ class SConsValues(optparse.Values):
                     # is not available.
                     raise AttributeError(attr)
 
-
+    # keep this list in sync with the SetOption doc in SCons/Script/Main.xml
+    # search for UPDATE_SETOPTION_DOCS there.
     settable = [
         'clean',
         'diskcheck',
         'duplicate',
+        'experimental',
+        'hash_chunksize',
+        'hash_format',
         'help',
         'implicit_cache',
+        'implicit_deps_changed',
+        'implicit_deps_unchanged',
         'max_drift',
         'md5_chunksize',
         'no_exec',
+        'no_progress',
         'num_jobs',
         'random',
+        'silent',
         'stack_size',
         'warn',
-        'silent'
+
+        # TODO: Remove these once we update the AddOption() API to allow setting
+        #       added flag as setable.
+        # Requested setable flag in : https://github.com/SCons/scons/issues/3983
+        # From experimental ninja
+        'disable_execute_ninja',
+        'disable_ninja'
     ]
 
     def set_option(self, name, value):
-        """
-        Sets an option from an SConscript file.
-        """
-        if name not in self.settable:
-            raise SCons.Errors.UserError("This option is not settable from a SConscript file: %s"%name)
+        """Sets an option from an SConscript file.
 
+        Raises:
+            UserError: invalid or malformed option ("error in your script")
+        """
+
+        if name not in self.settable:
+            raise SCons.Errors.UserError(
+                "This option is not settable from a SConscript file: %s" % name
+            )
+
+        # the following are for options that need some extra processing
         if name == 'num_jobs':
             try:
                 value = int(value)
                 if value < 1:
                     raise ValueError
             except ValueError:
-                raise SCons.Errors.UserError("A positive integer is required: %s"%repr(value))
+                raise SCons.Errors.UserError("A positive integer is required: %s" % repr(value))
         elif name == 'max_drift':
             try:
                 value = int(value)
             except ValueError:
-                raise SCons.Errors.UserError("An integer is required: %s"%repr(value))
+                raise SCons.Errors.UserError(
+                    "An integer is required: %s" % repr(value))
         elif name == 'duplicate':
             try:
                 value = str(value)
             except ValueError:
-                raise SCons.Errors.UserError("A string is required: %s"%repr(value))
+                raise SCons.Errors.UserError(
+                    "A string is required: %s" % repr(value))
             if value not in SCons.Node.FS.Valid_Duplicates:
-                raise SCons.Errors.UserError("Not a valid duplication style: %s" % value)
+                raise SCons.Errors.UserError(
+                    "Not a valid duplication style: %s" % value
+                )
             # Set the duplicate style right away so it can affect linking
             # of SConscript files.
             SCons.Node.FS.set_duplicate(value)
@@ -176,7 +196,8 @@ class SConsValues(optparse.Values):
             try:
                 value = diskcheck_convert(value)
             except ValueError as v:
-                raise SCons.Errors.UserError("Not a valid diskcheck value: %s"%v)
+                raise SCons.Errors.UserError(
+                    "Not a valid diskcheck value: %s" % v)
             if 'diskcheck' not in self.__dict__:
                 # No --diskcheck= option was specified on the command line.
                 # Set this right away so it can affect the rest of the
@@ -186,17 +207,29 @@ class SConsValues(optparse.Values):
             try:
                 value = int(value)
             except ValueError:
-                raise SCons.Errors.UserError("An integer is required: %s"%repr(value))
-        elif name == 'md5_chunksize':
+                raise SCons.Errors.UserError(
+                    "An integer is required: %s" % repr(value))
+        elif name in ('md5_chunksize', 'hash_chunksize'):
             try:
                 value = int(value)
             except ValueError:
-                raise SCons.Errors.UserError("An integer is required: %s"%repr(value))
+                raise SCons.Errors.UserError(
+                    "An integer is required: %s" % repr(value))
+            name = 'md5_chunksize'  # for now, the old name is used
         elif name == 'warn':
             if SCons.Util.is_String(value):
                 value = [value]
             value = self.__SConscript_settings__.get(name, []) + value
             SCons.Warnings.process_warn_strings(value)
+        elif name == 'no_progress':
+            SCons.Script.Main.progress_display.set_mode(False)
+        elif name == 'experimental':
+            if SCons.Util.is_String(value):
+                value = [value]
+            value = self.__SConscript_settings__.get(name, []) + value
+        elif name in ('implicit_deps_changed', 'implicit_deps_unchanged'):
+            if value:
+                self.__SConscript_settings__['implicit_cache'] = True
 
         self.__SConscript_settings__[name] = value
 
@@ -226,9 +259,13 @@ class SConsOption(optparse.Option):
             fmt = "option %s: nargs='?' is incompatible with short options"
             raise SCons.Errors.UserError(fmt % self._short_opts[0])
 
-    CHECK_METHODS = optparse.Option.CHECK_METHODS + [_check_nargs_optional]
+    CHECK_METHODS = optparse.Option.CHECK_METHODS
+    if CHECK_METHODS is None:
+        CHECK_METHODS = []
+    CHECK_METHODS = CHECK_METHODS + [_check_nargs_optional]
     CONST_ACTIONS = optparse.Option.CONST_ACTIONS + optparse.Option.TYPED_ACTIONS
 
+
 class SConsOptionGroup(optparse.OptionGroup):
     """
     A subclass for SCons-specific option groups.
@@ -237,6 +274,7 @@ class SConsOptionGroup(optparse.OptionGroup):
     the group's help text flush left, underneath their own title but
     lined up with the normal "SCons Options".
     """
+
     def format_help(self, formatter):
         """
         Format an option group's help text, outdenting the title so it's
@@ -248,6 +286,7 @@ class SConsOptionGroup(optparse.OptionGroup):
         result = result + optparse.OptionContainer.format_help(self, formatter)
         return result
 
+
 class SConsOptionParser(optparse.OptionParser):
     preserve_unknown_options = False
 
@@ -269,8 +308,7 @@ class SConsOptionParser(optparse.OptionParser):
         """
         arg = rargs.pop(0)
 
-        # Value explicitly attached to arg?  Pretend it's the next
-        # argument.
+        # Value explicitly attached to arg?  Pretend it's the next argument.
         if "=" in arg:
             (opt, next_arg) = arg.split("=", 1)
             rargs.insert(0, next_arg)
@@ -280,7 +318,11 @@ class SConsOptionParser(optparse.OptionParser):
             had_explicit_value = False
 
         try:
-            opt = self._match_long_opt(opt)
+            if opt != self._match_long_opt(opt):
+                raise optparse.BadOptionError(
+                    "'%s'. Did you mean '%s'?"
+                    % (opt, self._match_long_opt(opt))
+                )
         except optparse.BadOptionError:
             if self.preserve_unknown_options:
                 # SCons-specific:  if requested, add unknown options to
@@ -306,7 +348,7 @@ class SConsOptionParser(optparse.OptionParser):
                     if not option.choices:
                         self.error(_("%s option requires an argument") % opt)
                     else:
-                        msg  = _("%s option requires an argument " % opt)
+                        msg = _("%s option requires an argument " % opt)
                         msg += _("(choose from %s)"
                                  % ', '.join(option.choices))
                         self.error(msg)
@@ -328,29 +370,27 @@ class SConsOptionParser(optparse.OptionParser):
         option.process(opt, value, values, self)
 
     def reparse_local_options(self):
-        """
-        Re-parse the leftover command-line options stored
-        in self.largs, so that any value overridden on the
-        command line is immediately available if the user turns
-        around and does a GetOption() right away.
+        """ Re-parse the leftover command-line options.
+
+        Parse options stored in `self.largs`, so that any value
+        overridden on the command line is immediately available
+        if the user turns around and does a :func:`GetOption` right away.
 
         We mimic the processing of the single args
-        in the original OptionParser._process_args(), but here we
-        allow exact matches for long-opts only (no partial
-        argument names!).
-
-        Else, this would lead to problems in add_local_option()
+        in the original OptionParser :func:`_process_args`, but here we
+        allow exact matches for long-opts only (no partial argument names!).
+        Otherwise there could be  problems in :func:`add_local_option`
         below. When called from there, we try to reparse the
         command-line arguments that
-          1. haven't been processed so far (self.largs), but
-          2. are possibly not added to the list of options yet.
 
-        So, when we only have a value for "--myargument" yet,
-        a command-line argument of "--myarg=test" would set it.
-        Responsible for this behaviour is the method
-        _match_long_opt(), which allows for partial matches of
-        the option name, as long as the common prefix appears to
-        be unique.
+        1. haven't been processed so far (`self.largs`), but
+        2. are possibly not added to the list of options yet.
+
+        So, when we only have a value for "--myargument" so far,
+        a command-line argument of "--myarg=test" would set it,
+        per the behaviour of :func:`_match_long_opt`,
+        which allows for partial matches of the option name,
+        as long as the common prefix appears to be unique.
         This would lead to further confusion, because we might want
         to add another option "--myarg" later on (see issue #2929).
 
@@ -396,7 +436,7 @@ class SConsOptionParser(optparse.OptionParser):
         """
         Adds a local option to the parser.
 
-        This is initiated by a SetOption() call to add a user-defined
+        This is initiated by an AddOption() call to add a user-defined
         command-line option.  We add the option to a separate option
         group for the local options, creating the group if necessary.
         """
@@ -423,6 +463,7 @@ class SConsOptionParser(optparse.OptionParser):
 
         return result
 
+
 class SConsIndentedHelpFormatter(optparse.IndentedHelpFormatter):
     def format_usage(self, usage):
         return "usage: %s\n" % usage
@@ -471,8 +512,8 @@ class SConsIndentedHelpFormatter(optparse.IndentedHelpFormatter):
         opt_width = self.help_position - self.current_indent - 2
         if len(opts) > opt_width:
             wrapper = textwrap.TextWrapper(width=self.width,
-                                           initial_indent = '  ',
-                                           subsequent_indent = '  ')
+                                           initial_indent='  ',
+                                           subsequent_indent='  ')
             wrapper.wordsep_re = no_hyphen_re
             opts = wrapper.fill(opts) + '\n'
             indent_first = self.help_position
@@ -486,7 +527,7 @@ class SConsIndentedHelpFormatter(optparse.IndentedHelpFormatter):
 
             # SCons:  indent every line of the help text but the first.
             wrapper = textwrap.TextWrapper(width=self.help_width,
-                                           subsequent_indent = '  ')
+                                           subsequent_indent='  ')
             wrapper.wordsep_re = no_hyphen_re
             help_lines = wrapper.wrap(help_text)
             result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
@@ -496,6 +537,7 @@ class SConsIndentedHelpFormatter(optparse.IndentedHelpFormatter):
             result.append("\n")
         return "".join(result)
 
+
 def Parser(version):
     """
     Returns an options parser object initialized with the standard
@@ -585,15 +627,15 @@ def Parser(version):
 
     def opt_invalid(group, value, options):
         """report an invalid option from a group"""
-        errmsg  = "`%s' is not a valid %s option type, try:\n" % (value, group)
+        errmsg = "`%s' is not a valid %s option type, try:\n" % (value, group)
         return errmsg + "    %s" % ", ".join(options)
 
     def opt_invalid_rm(group, value, msg):
         """report an invalid option from a group: recognized but removed"""
-        errmsg  = "`%s' is not a valid %s option type " % (value, group)
+        errmsg = "`%s' is not a valid %s option type " % (value, group)
         return errmsg + msg
 
-    config_options = ["auto", "force" ,"cache"]
+    config_options = ["auto", "force", "cache"]
 
     opt_config_help = "Controls Configure subsystem: %s." \
                       % ", ".join(config_options)
@@ -601,7 +643,7 @@ def Parser(version):
     op.add_option('--config',
                   nargs=1, choices=config_options,
                   dest="config", default="auto",
-                  help = opt_config_help,
+                  help=opt_config_help,
                   metavar="MODE")
 
     op.add_option('-D',
@@ -613,10 +655,10 @@ def Parser(version):
     deprecated_debug_options = {}
 
     removed_debug_options = {
-        "dtree"         : '; please use --tree=derived instead',
-        "nomemoizer"    : '; there is no replacement',
-        "stree"         : '; please use --tree=all,status instead',
-        "tree"          : '; please use --tree=all instead',
+        "dtree": '; please use --tree=derived instead',
+        "nomemoizer": '; there is no replacement',
+        "stree": '; please use --tree=all,status instead',
+        "tree": '; please use --tree=all instead',
     }
 
     debug_options = ["count", "duplicate", "explain", "findlibs",
@@ -645,7 +687,8 @@ def Parser(version):
                 msg = removed_debug_options[value]
                 raise OptionValueError(opt_invalid_rm('debug', value, msg))
             else:
-                raise OptionValueError(opt_invalid('debug', value, debug_options))
+                raise OptionValueError(opt_invalid(
+                    'debug', value, debug_options))
 
     opt_debug_help = "Print various types of debugging information: %s." \
                      % ", ".join(debug_options)
@@ -673,7 +716,7 @@ def Parser(version):
     def opt_duplicate(option, opt, value, parser):
         if value not in SCons.Node.FS.Valid_Duplicates:
             raise OptionValueError(opt_invalid('duplication', value,
-                                              SCons.Node.FS.Valid_Duplicates))
+                                               SCons.Node.FS.Valid_Duplicates))
         setattr(parser.values, option.dest, value)
         # Set the duplicate style right away so it can affect linking
         # of SConscript files.
@@ -690,9 +733,42 @@ def Parser(version):
 
     if not SCons.Platform.virtualenv.virtualenv_enabled_by_default:
         op.add_option('--enable-virtualenv',
-                     dest="enable_virtualenv",
-                     action="store_true",
-                     help="Import certain virtualenv variables to SCons")
+                      dest="enable_virtualenv",
+                      action="store_true",
+                      help="Import certain virtualenv variables to SCons")
+
+    def experimental_callback(option, opt, value, parser):
+        experimental = getattr(parser.values, option.dest)
+
+        if ',' in value:
+            value = value.split(',')
+        else:
+            value = [value]
+
+        for v in value:
+            if v == 'none':
+                experimental = set()
+            elif v == 'all':
+                experimental = experimental_features
+            elif v not in experimental_features:
+                raise OptionValueError(
+                    "option --experimental: invalid choice: '%s' "
+                    "(choose from 'all','none',%s)"
+                    % (v, ','.join(["'%s'" % e for e in sorted(experimental_features)]))
+                )
+            else:
+                experimental |= {v}
+
+        setattr(parser.values, option.dest, experimental)
+
+    op.add_option('--experimental',
+                  dest='experimental',
+                  action='callback',
+                  default=set(),  # empty set
+                  type='str',
+                  # choices=experimental_options+experimental_features,
+                  callback=experimental_callback,
+                  help='Enable experimental features')
 
     op.add_option('-f', '--file', '--makefile', '--sconstruct',
                   nargs=1, type="string",
@@ -709,6 +785,27 @@ def Parser(version):
                   action="help",
                   help="Print this message and exit.")
 
+    def warn_md5_chunksize_deprecated(option, opt, value, parser):
+        if opt == '--md5-chunksize':
+            SCons.Warnings.warn(SCons.Warnings.DeprecatedWarning,
+                                "Parameter %s is deprecated. Use "
+                                "--hash-chunksize instead." % opt)
+
+        setattr(parser.values, option.dest, value)
+
+    op.add_option('--hash-chunksize', '--md5-chunksize',
+                  nargs=1, type="int",
+                  dest='md5_chunksize', default=SCons.Node.FS.File.hash_chunksize,
+                  action="callback",
+                  help="Set chunk-size for hash signature computation to N kilobytes.",
+                  callback=warn_md5_chunksize_deprecated,
+                  metavar="N")
+
+    op.add_option('--hash-format',
+                  dest='hash_format',
+                  action='store',
+                  help='Hash format (e.g. md5, sha1, or sha256).')
+
     op.add_option('-i', '--ignore-errors',
                   dest='ignore_errors', default=False,
                   action="store_true",
@@ -722,9 +819,9 @@ def Parser(version):
                   metavar="DIR")
 
     op.add_option('--ignore-virtualenv',
-                 dest="ignore_virtualenv",
-                 action="store_true",
-                 help="Do not import virtualenv variables to SCons")
+                  dest="ignore_virtualenv",
+                  action="store_true",
+                  help="Do not import virtualenv variables to SCons")
 
     op.add_option('--implicit-cache',
                   dest='implicit_cache', default=False,
@@ -769,21 +866,14 @@ def Parser(version):
                   help="Set maximum system clock drift to N seconds.",
                   metavar="N")
 
-    op.add_option('--md5-chunksize',
-                  nargs=1, type="int",
-                  dest='md5_chunksize', default=SCons.Node.FS.File.md5_chunksize,
-                  action="store",
-                  help="Set chunk-size for MD5 signature computation to N kilobytes.",
-                  metavar="N")
-
     op.add_option('-n', '--no-exec', '--just-print', '--dry-run', '--recon',
                   dest='no_exec', default=False,
                   action="store_true",
                   help="Don't build; just print commands.")
 
     op.add_option('--no-site-dir',
-                  dest='no_site_dir', default=False,
-                  action="store_true",
+                  dest='site_dir',
+                  action="store_false",
                   help="Don't search or use the usual site_scons dir.")
 
     op.add_option('--profile',
@@ -834,10 +924,9 @@ def Parser(version):
                   help="Trace Node evaluation to FILE.",
                   metavar="FILE")
 
-    tree_options = ["all", "derived", "prune", "status"]
+    tree_options = ["all", "derived", "prune", "status", "linedraw"]
 
     def opt_tree(option, opt, value, parser, tree_options=tree_options):
-        from . import Main
         tp = Main.TreePrinter()
         for o in value.split(','):
             if o == 'all':
@@ -848,6 +937,8 @@ def Parser(version):
                 tp.prune = True
             elif o == 'status':
                 tp.status = True
+            elif o == 'linedraw':
+                tp.sLineDraw = True
             else:
                 raise OptionValueError(opt_invalid('--tree', o, tree_options))
         parser.values.tree_printers.append(tp)
@@ -899,7 +990,6 @@ def Parser(version):
                   action="append",
                   help="Search REPOSITORY for source and target files.")
 
-
     # Options from Make and Cons classic that we do not yet support,
     # but which we may support someday and whose (potential) meanings
     # we don't want to change.  These all get a "the -X option is not
